"""Simple HTTP server for demonstrating the http plug in"""

from BaseHTTPServer import BaseHTTPRequestHandler
import cgi
import sys

class DemoHttpServer(BaseHTTPRequestHandler):
    """Keep a simple dictionary of request paths.  Return (0,0) if hasn't
    been set before."""
    
    VALUE_DICT = {}
    
    def do_GET(self): # pylint: disable=C0103
        """Return the value recorded for the path"""
        value = DemoHttpServer.VALUE_DICT.get(self.path, "0,0")
        
        self.send_response(200)
        self.end_headers()
        self.wfile.write(value)
        return

    def do_POST(self): # pylint: disable=C0103
        """Add to VALUE_DICT the key of the request path and the value of
        whatever is in the "block" parameter."""
        form = cgi.FieldStorage(
            fp=self.rfile, 
            headers=self.headers,
            environ={'REQUEST_METHOD':'POST',
                     'CONTENT_TYPE':self.headers['Content-Type'],
                     })

        # Begin the response
        self.send_response(200)
        self.end_headers()
        
        value = form["block"].value
        DemoHttpServer.VALUE_DICT[self.path] = value
        
        self.wfile.write("Value set = " + value)
        
        return        
        
        
if __name__ == '__main__':
    SERVERNAME = "localhost"
    PORT = 8080
    if len(sys.argv) > 1:
        SERVERNAME = sys.argv[1]
    if len(sys.argv) > 2:
        PORT = int(sys.argv[2])
    from BaseHTTPServer import HTTPServer
    SERVER = HTTPServer((SERVERNAME, PORT), DemoHttpServer)
    print 'Starting server, use <Ctrl-C> to stop'
    SERVER.serve_forever()


"""This module is an example of a plugin implementation and also used for
unittesting.
 """

from rube_mc_pi.rube import Source, Target

class MockSource(Source): #pylint: disable=R0903
    """Stand in for a real source object, counting times called and updating
    as requested by the test methods.
    """
    def __init__(self, attribs):
        super(MockSource, self).__init__()
        self.was_poll_state_called = False
        self.state = attribs.get("state", 1)
        self.query_count = attribs.get("query_count", 0)
        self.loops_before_stop = None
        self.state_changes = {}
        self.was_constructor_called = True

    def poll_state(self):
        self.was_poll_state_called = True
        self.query_count = self.query_count + 1
        if self.loops_before_stop != None:
            # Need to take account of the initial poll_state before the 1st loop
            if self.query_count > self.loops_before_stop + 1:
                raise KeyboardInterrupt()

        if self.state_changes.has_key(self.query_count):
            self.state = self.state_changes[self.query_count]

        return self.state


class MockTarget(Target): #pylint: disable=R0903
    """Stand in for a real target object.  Takes record of calls made to update
    """
    def __init__(self, attribs):
        super(MockTarget, self).__init__()
        self.was_update_state_called = False
        self.state_log = []
        self.name = attribs.get("name", "")
        self.was_constructor_called = True


    def update_state(self, block):
        self.state_log.append(block)
        self.was_update_state_called = True







"""
Adaptor for scratch using the scratchpy interface

To function as a source, the scratch program should just broadcast a block state
e.g. broadcast 41,0

To function as a target, the scratch program should listen for broadcasts saying
either BLOCK_PRESENT or BLOCK_ABSENT.  Implementation of individual block types
has not yet been implemented

The JSON config looks like this:

    { "source": {
        "type": "scratchplug",
        "server_address": "localhost",
        "server_port": 42001 
        },
        
      "target":  {
        "type": "scratchplug",
        "server_address": "localhost",
        "server_port": 42001 
        }
    }

"""

import rube_mc_pi.mcpi.block as block
from multiprocessing import Process, Queue
from Queue import Empty
import rube_mc_pi.rube as rube
import scratch as scratch_ex


class ScratchplugSource(rube.Source): #pylint: disable=R0903
    """Wait for a broadcast from scratch showing that the block has changed
    The standard scratch.receive() method blocks until a message is received.
    This implementation wraps the call in a multiprocessing loop to give up
    if nothing is received before a time out
    """
    def __init__(self, attribs):
        super(ScratchplugSource, self).__init__()
        self.scratch_link = ScratchLink(attribs)
        self.last_block_state = block.AIR
        self.from_scratch = None

    def receive_from_scratch(self, queue):
        """Look for a "x,y" broadcast and convert to a block.
        If a broadcast is received that doesn't parse into x,y then leave the
        state unchanged.
        """
        self.from_scratch = None
        received = self.scratch_link.scratch_connection.receive()[1]
        try:
            id_, data = received.split(",")
            _block = block.Block(int(id_), int(data))
        except ValueError:
            #Assume that the broadcast wasn't intended as a block change
            print("Unparsable broadcast:" + received)
            _block = None
            
        if _block:
            queue.put(_block)
        
    def poll_state(self):
        """Assumes that scratch will broadcast "x,y" to represent the
        block state.        
        """
        queue = Queue()
        proc = Process(target=self.receive_from_scratch, args=(queue,))
        proc.start()
     
        timout_secs = 3
        proc.join(timout_secs)

        if proc.is_alive():
            proc.terminate()
            proc.join()
            
        try:
            _block = queue.get_nowait()
            self.last_block_state = _block
        except Empty:
            pass
        
        return self.last_block_state
    
    

class ScratchplugTarget(rube.Target): #pylint: disable=R0903
    """Send a broadcast to scratch with either "BLOCK_PRESENT" or "BLOCK_ABSENT"
    """
    def __init__(self, attribs):
        super(ScratchplugTarget, self).__init__()
        self.scratch_link = ScratchLink(attribs)

    def update_state(self, block_):
        if block_ != block.AIR:
            self.scratch_link.scratch_connection.broadcast("BLOCK_PRESENT")
        else:
            self.scratch_link.scratch_connection.broadcast("BLOCK_ABSENT")

                                    
class ScratchLink(object):
    """
    Responsible for storing the state of the link to a server
    """
                                       
    def __init__(self, attribs):
        self.server_address = attribs["server_address"]
        self.server_port = attribs.get("server_port", 42001)
        self.scratch_connection = scratch_ex.Scratch(self.server_address,
                                                     port=self.server_port)   
   


"""Code to manage the Rube Goldberg project for the Manchester CoderDojo """

import collections
import importlib
import json
import time


class RubeController(object):
    """"Responsible for controlling the interactions between components in
    the Rube Goldberg machine.
    """

    def __init__(self, config, min_loop_duration_ms=100):
        """Set up a controller using the passed config.

        The config is a tuple of (source, target) pairs.
        """
        self.config = config
        self.min_loop_duration_ms = min_loop_duration_ms

    def run_event_loop(self):
        """Run the event loop continuously looking for state changes"""

        self.record_initial_states()
        while True:
            start_time = time.time()
            self.update_all_once()
            loop_run_time_ms = (time.time() - start_time) * 1000
            if loop_run_time_ms < self.min_loop_duration_ms:
                sleep_needed = (self.min_loop_duration_ms - loop_run_time_ms) \
                               /1000
                time.sleep(sleep_needed)

    def record_initial_states(self):
        """Record the initial state but don't trigger any update"""

        for (source, target) in self.config:
            target.last_state_update = source.poll_state()

    def update_all_once(self):
        """Run through all the config pairs once checking state and updating """

        for (source, target) in self.config:
            new_state = source.poll_state()
            if target.last_state_update != new_state:
                target.update_state(new_state)
                target.last_state_update = new_state

class Source(object): #pylint: disable=R0903
    """A source object represents the part of the machine that will trigger the
    next step when it completes.

    A source adaptor is responsible for responding to the poll_state() method by
    returning the current state of the source part of the machine.

    """

    def poll_state(self):
        """Ask a source component for its current state"""
        raise NotImplementedError(
        "It is for concrete implementations of this class to fill in the" \
        " poll_state() method.")


class Target(object): #pylint: disable=R0903
    """
    A target object represents the part of the machine that should be triggered
    to start of the next step running.

    A target adaptor is responsible for responding to the update_state(block)
    method by taking appropriate actions to update its state.

    """

    def __init__(self):
        self.last_state_update = "test"

    def update_state(self, block):
        """Ask the target component to set its state to the passed Minecraft
        Block object value."""
        raise NotImplementedError("It is for concrete implementations of this" \
            "class to fill in the update_state(block) method.")



ConfigPair = collections.namedtuple("ConfigPair", "source target") # pylint: disable=C0103, C0301

class ConfigJsonParser(object):
    """Responsible for reading in JSON config and converting it to a config
    object that can set up the RubeController

    """

    @staticmethod
    def make_instance_from_dict(attrib_dict, source_or_target):
        """Build the objects based on the passed
        dictionary of attributes.
        The 'type' attribute defines which class of object to load

        """
        type_ = attrib_dict["type"]
        core_attribs = {name: value for (name, value)
                            in attrib_dict.items()
                            if name != "type"}

        return ConfigJsonParser.make_instance(type_,
                                              source_or_target,
                                              core_attribs)


    @staticmethod
    def make_instance(module_name, type_, attribute_dict):
        """Try to load the module of the given name, find a class in
        that module of type module_nametype, e.g. ExampleSource, and
        then return a new instance with all the attributes set
        as per the passed dictionary
        """
        name = "rube_mc_pi." + module_name
        module = importlib.import_module(name)
        class_name = module_name.capitalize() + type_.capitalize()
        class_ = getattr(module, class_name)
        instance = class_(attribute_dict)

        return instance


    @staticmethod
    def parse(json_string):
        """Take the json and return (source, target) pairs ready for use"""

        config = []
        jsonparse = json.loads(json_string)
        for config_pair in jsonparse:
            source = ConfigJsonParser.make_instance_from_dict(
                                                config_pair["source"], "source")
            target = ConfigJsonParser.make_instance_from_dict(
                                                config_pair["target"], "target")
            config.append(ConfigPair(source=source, target=target))

        return config

if __name__ == "__main__":
    JSON_CONFIG = ""
    with open("config.json", "r") as f:
        JSON_CONFIG = f.read()
    
    CONFIG = ConfigJsonParser.parse(JSON_CONFIG)
    CONTROLLER = RubeController(CONFIG)
    CONTROLLER.run_event_loop()



"""
Adaptor for Arduino Firmata protocol.

NOTE:  This plugin is experimental as there has not been time to test it with
real ardunios at the time of writing.  The plugin is currently a simple port 
of the GPIO code to use the Firmata API.

The JSON config looks like this:

    { "source": {
        "type": "firmata",
        "board": "/dev/tty.usbserial-A6008rIF",
        "pin": 18,
        "low_state_block": [0, 0],
        "high_state_block": [45, 0]
        },
        
      "target":  {
        "type": "firmata",
        "board": "/dev/tty.usbserial-A6008rIF",
        "pin": 12
        }
    }
"""

from rube_mc_pi.mcpi import block
from rube_mc_pi.mcpi.block import Block
import pyfirmata
import rube_mc_pi.rube as rube
import time



class FirmataSource(rube.Source): #pylint: disable=R0903
    """
        Use the input from the firmata pin.
		The low_state_block is the block to report if the pin in low
		The high_state_block is what to report if the pin is high
    """
    
    
    def __init__(self, attribs):
        super(FirmataSource, self).__init__()
        self.low_state_block = Block(attribs["low_state_block"][0], 
                                     attribs["low_state_block"][1])
        self.high_state_block = Block(attribs["high_state_block"][0],
                                      attribs["high_state_block"][1])
        self.board_name = attribs["board"]
        self.board = pyfirmata.Arduino(self.board_name)
        iterator = pyfirmata.util.Iterator(self.board)
        iterator.start()
        time.sleep(1)
        self.pin_number = attribs["pin"]
        self.pin = self.board.get_pin("d:%d:i" % self.pin_number)
        time.sleep(1)
        self.pin.enable_reporting()
        time.sleep(1)

    def poll_state(self):
        if self.pin.read():
            return self.high_state_block
        else:
            return self.low_state_block

class FirmataTarget(rube.Target): #pylint: disable=R0903
    """
        Simply if the state passed is Block.AIR (0,0) the turn output low.
        Any other block type put the output high
    """
            
        
    def __init__(self, attribs):
        super(FirmataTarget, self).__init__()
        self.board_name = attribs["board"]
        self.board = pyfirmata.Arduino(self.board_name)
        time.sleep(1)
        self.pin_number = attribs["pin"]
        self.pin = self.board.get_pin("d:%d:o" % self.pin_number)
        time.sleep(1)
        

    def update_state(self, new_state):
        if new_state == block.AIR:
            self.pin.write(0)
        else:
            self.pin.write(1)



"""
Adaptor for Raspberry Pi GPIO plugs.

The JSON config looks like this:

    { "source": {
        "type": "gpio",
        "pin": 18,
        "low_state_block": [0, 0],
        "high_state_block": [45, 0]
        },
        
      "target":  {
        "type": "gpio",
        "pin": 12
        }
    }
"""

from rube_mc_pi.mcpi import block
from rube_mc_pi.mcpi.block import Block
import RPi.GPIO as GPIO
import rube_mc_pi.rube as rube



class GpioSource(rube.Source): #pylint: disable=R0903
    """
        Use the input from the raspberry pi GPIO pin.
		The low_state_block is the block to report if the pin in low
		The high_state_block is what to report if the pin is high
    """
    
    @staticmethod
    def gpio_in_setup(pin):
        """Set the pin up for input"""
        GPIO.setmode(GPIO.BOARD)
        GPIO.setup(pin, GPIO.IN, pull_up_down=GPIO.PUD_UP)

    
    def __init__(self, attribs):
        super(GpioSource, self).__init__()
        self.pin = attribs["pin"]
        self.low_state_block = Block(attribs["low_state_block"][0], 
                                     attribs["low_state_block"][1])
        self.high_state_block = Block(attribs["high_state_block"][0],
                                      attribs["high_state_block"][1])
        GpioSource.gpio_in_setup(self.pin)

    def poll_state(self):
        if GPIO.input(self.pin):
            return self.low_state_block
        else:
            return self.high_state_block

class GpioTarget(rube.Target): #pylint: disable=R0903
    """
        Simply if the state pass is Block.AIR (0,0) the turn output low.
        Any other block type put the output high
    """
    
    @staticmethod
    def gpio_out_setup(pin):
        """Set GPIO up for output and initialise it to low"""
        GPIO.setmode(GPIO.BOARD)
        GPIO.setup(pin, GPIO.OUT)
        GPIO.output(pin, GPIO.LOW)
        
        
    def __init__(self, attribs):
        super(GpioTarget, self).__init__()
        self.pin = attribs["pin"]
        GpioTarget.gpio_out_setup(self.pin)

    def update_state(self, new_state):
        if new_state == block.AIR:
            GPIO.output(self.pin, GPIO.LOW)
        else:
            GPIO.output(self.pin, GPIO.HIGH)



"""
Adaptor for minecraft using Bukkit and the Raspberry Juice plugin

The JSON looks like this:

{ "source": {
        "type": "minecraft",
        "coords_x": 1,
        "coords_y": 10,
        "coords_z": 3,
        "server_address": "localhost"
        },
        
      "target":  {
        "type": "minecraft",
        "coords_x": 2,
        "coords_y": 10,
        "coords_z": 3,
        "server_address": "localhost"
        }


"""


from rube_mc_pi.mcpi.minecraft import Minecraft
import rube_mc_pi.mcpi.block as block
import rube_mc_pi.rube as rube


class MinecraftSource(rube.Source): #pylint: disable=R0903
    """Monitor a minecraft server for the block state at given coordinates"""
    def __init__(self, attribs):
        super(MinecraftSource, self).__init__()
        self.mc_link = MinecraftLink(attribs)

    def poll_state(self):
        """Return the block value from the connected Minecraft server"""
        block_ = self.mc_link.world_connection.getBlock(self.mc_link.coords_x,
                                                        self.mc_link.coords_y,
                                                        self.mc_link.coords_z)
        
        return block_

class MinecraftTarget(rube.Target): #pylint: disable=R0903
    """Set the block on the server to the specified value"""
    def __init__(self, attribs):
        super(MinecraftTarget, self).__init__()
        self.mc_link = MinecraftLink(attribs)

    def update_state(self, block_):
        self.mc_link.world_connection.setBlock(self.mc_link.coords_x,
                                               self.mc_link.coords_y,
                                               self.mc_link.coords_z,
                                               block_)

                                       
class MinecraftLink(object):   #pylint: disable=R0903
    """
    Responsible for storing the state of the link to a server and the location
    of the block that is to be monitored or updated.

    The link object should manage connections so that only one connection is created
    to each server.  This connection is shared between all instances that need it.
    """

    connection_pool = {}
    
    @staticmethod
    def reuse_or_create_connection(address,  port):
        """Make sure only one connections is opened to each server"""
        if (address,  port) not in MinecraftLink.connection_pool:
            MinecraftLink.connection_pool[(address,  port)] =  \
                            Minecraft.create(address, port)
        
        return MinecraftLink.connection_pool[(address,  port)]
            
        
    def __init__(self, attribs):
        self.server_address = attribs["server_address"]
        self.server_port = attribs.get("server_port", 4711)
        self.coords_x = attribs["coords_x"]
        self.coords_y = attribs["coords_y"]
        self.coords_z = attribs["coords_z"]
        self.world_connection = \
            MinecraftLink.reuse_or_create_connection(self.server_address,
                                                                            self.server_port)


if __name__ == "__main__":
    JSON_CONFIG = ""
    with open("config.json", "r") as f:
        JSON_CONFIG = f.read()
    
    CONFIG = rube.ConfigJsonParser.parse(JSON_CONFIG)
    SOURCE = CONFIG[0][0]
    SOURCE.mc_link.world_connection.setBlock(1,
                                             10,
                                             3,
                                             block.TNT)
    CONTROLLER = rube.RubeController(CONFIG)
    CONTROLLER.run_event_loop()



"""Simple script to print out the (x,y,z) location of the current player on
a minecraft server


usage:  python playerloc.py [server-name] [port]


"""

import rube_mc_pi.mcpi.minecraft as minecraft
import sys


if __name__ == "__main__":
    SERVERNAME = "localhost"
    PORT = 4711
    if len(sys.argv) > 1:
        SERVERNAME = sys.argv[1]
    if len(sys.argv) > 2:
        PORT = int(sys.argv[2])
    print("Opening world '%s:%d'" % (SERVERNAME, PORT))
    WORLD = minecraft.Minecraft.create(SERVERNAME, PORT)
    print("Connected to world")
    POSITION = WORLD.player.getPos()
    print("Player is at (%d, %d, %d)" % (POSITION.x, POSITION.y, POSITION.z))



"""
Adaptor for communicating status to and from a http connection.  
The file format is a single line in a file with id,data,  e.g. 123,45

Example JSON config:

     { "source": {
        "type": "http",
        "address": "http://test.com/resource_status"
        },
        
      "target":  {
        "type": "http",
        "address": "http://test.com/update_handler"
        }
    }
"""

from rube_mc_pi.mcpi.block import Block
import rube_mc_pi.rube as rube
import urllib
import urllib2

class HttpSource(rube.Source): #pylint: disable=R0903
    """Obtain a status update by sending a GET request to an address"""
    def __init__(self, attribs):
        super(HttpSource, self).__init__()
        self.address = attribs["address"]

        
    def poll_state(self):
        response = urllib2.urlopen(self.address).read()
        parsed = response.split(",")
        return Block(id=int(parsed[0]), data=int(parsed[1]))
        
class HttpTarget(rube.Target): #pylint: disable=R0903
    """Update the status by sending a POST request with block=new_value"""

    @classmethod
    def build_post_request(cls, block_value):
        """Return the tuple of data, headers suitable for sending to a
        HTTP POST request to update the server to the new block_value"""
        
        data = cls.build_post_data(block_value)
        headers = cls.build_post_headers(data) 
        
        return data, headers

    
    @classmethod
    def build_post_data(cls, block_value):
        """Create the data part of the HTTP POST request to update the block"""
        return urllib.urlencode({'block': block_value})
    
    @classmethod
    def build_post_headers(cls, post_data):
        """Create the HTTP headers to send matching the post_data.
        Assumes that the post_data has already been urlencoded and will not
        be changed again before sending to the server"""
        
        return {"Content-Length": str(len(post_data))}        


        
    def __init__(self, attribs):
        super(HttpTarget, self).__init__()
        self.address = attribs["address"]

    def update_state(self, block_):
        """Send a POST request to update the server"""
        data_to_write = "%d,%d" % (block_.id, block_.data)
        request = urllib2.Request(self.address,
                                  *self.build_post_request(data_to_write))
        response = urllib2.urlopen(request)
        the_page = response.read()
        print("Sent http post.  Response = " + the_page)
        
          
